/// <reference types="node" />
/// <reference types="node" />
import EventEmitter from "events";
import { ClientResponseType } from "../client";
import { BuilderInstance } from "../builder";
import { CommandResponseDetails } from "../managers";
import { CacheOptionsType, CacheAsyncStorageType, CacheStorageType, getCacheEvents, CacheValueType } from "./";
import { CommandDump, CommandInstance } from "../command";
/**
 * Cache class handles the data exchange with the dispatchers.
 *
 * @note
 * Keys used to save the values are created dynamically on the Command class
 *
 */
export declare class Cache {
    builder: BuilderInstance;
    options?: CacheOptionsType;
    emitter: EventEmitter;
    events: ReturnType<typeof getCacheEvents>;
    storage: CacheStorageType;
    lazyStorage?: CacheAsyncStorageType;
    clearKey: string;
    garbageCollectors: Map<string, NodeJS.Timeout>;
    private logger;
    constructor(builder: BuilderInstance, options?: CacheOptionsType);
    /**
     * Set the cache data to the storage
     * @param command
     * @param response
     * @param details
     * @returns
     */
    set: <Response_1, Error_1>(command: CommandInstance | CommandDump<CommandInstance>, response: ClientResponseType<Response_1, Error_1>, details: CommandResponseDetails) => void;
    /**
     * Get particular record from storage by cacheKey. It will trigger lazyStorage to emit lazy load event for reading it's data.
     * @param cacheKey
     * @returns
     */
    get: <Response_1, Error_1>(cacheKey: string) => CacheValueType<Response_1, Error_1>;
    /**
     * Get sync storage keys, lazyStorage keys will not be included
     * @returns
     */
    keys: () => string[];
    /**
     * Delete record from storages and trigger revalidation
     * @param cacheKey
     */
    delete: (cacheKey: string) => void;
    /**
     * Revalidate cache by cacheKey or partial matching with RegExp
     * @param cacheKey
     */
    revalidate: (cacheKey: string | RegExp) => Promise<void>;
    /**
     * Used to receive data from lazy storage
     * @param cacheKey
     */
    getLazyResource: <Response_1, Error_1>(cacheKey: string) => Promise<CacheValueType<Response_1, Error_1>>;
    /**
     * Used to receive keys from sync storage and lazy storage
     * @param cacheKey
     */
    getLazyKeys: () => Promise<string[]>;
    /**
     * Schedule garbage collection for given key
     * @param cacheKey
     * @returns
     */
    scheduleGarbageCollector: (cacheKey: string) => Promise<void>;
    /**
     * Clear cache storages
     */
    clear: () => Promise<void>;
}
